;nyquist plug-in
;type analyze
;name "Click Finder..."
;action "Calculating..."

;control control-separation "Block length (ms)\n(maximum click length and minimum\nseparation of equal frequency clicks)" real "" 10 0 100
;control control-steps-per-block "Steps per block\n(higher is slower, improves precision)" int "" 2 1 100
;control control-relative-threshold "Peak-to-background threshold (dB)\n(lower increases sensitivity)" real "" 3 0.5 42
;control control-absolute-threshold "Peak amplitude threshold (dB)\n(lower increases sensitivity)" real "" -50 -100 0
;control control-frequency-floor "Minimum test frequency (Hz)" real "" 250 0 22050
;control control-frequency-ceiling "Maximum test frequency (Hz)" real "" 10000 0 22050
;control control-frequency-step "Test frequency increment (Hz)" real "" 250 0 4000

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; utilities

(defmacro inc (x) `(setf ,x (1+ ,x)))

(defmacro add-to (x val) `(setf ,x (+ ,val ,x)))

(defmacro update-op (op x val)
  (let ((temp1 (gensym))
	(temp2 (gensym)))
    `(let ((,temp1 ,val)
	   (,temp2 ,x))
       (setf ,x (if ,temp2 (,op ,temp2 ,temp1) ,temp1)))))
(defmacro update-max (x val) `(update-op max ,x ,val))
(defmacro update-min (x val) `(update-op min ,x ,val))

(defmacro push-back (val lis) `(setf ,lis (cdr (rplacd ,lis (cons ,val nil)))))

;;; I was so annoyed at the lack of a loop that does not test until after the
;;; first iteration...
(defun loop1
  (head bindings termination body-forms)
  (let ((initialized (gensym)))
    (flet
	((expand-binding
	  (binding)
	  (if (atom binding)
	      binding
	      `(,(car binding) nil ,@(cdr binding)))))
      (let ((expanded-bindings (mapcar #'expand-binding bindings)))
	`(,head
	  (,initialized ,@expanded-bindings)
	  ((and ,initialized ,(car termination)) ,@(cdr termination))
	  (cond (,initialized ,@body-forms)
		(t (setf ,initialized t))))))))
(defmacro do1 (bindings termination &rest body-forms)
  (loop1 'do bindings termination body-forms))
(defmacro do1* (bindings termination &rest body-forms)
  (loop1 'do* bindings termination body-forms))

(defun loop-build-list
  (head bindings test body-forms)
  (let ((lis (gensym))
	(tail (gensym)))
    ;; put the loop in a block so that body-forms
    ;; can use return to stop accumulating
    `(let* ((,lis (list nil))
	    (,tail ,lis))
       (block nil (,head ,bindings
			 (,test)
			 (push-back (progn ,@body-forms) ,tail)))
       (cdr ,lis))))
(defmacro do-build-list (bindings test &rest body-forms)
  (loop-build-list 'do bindings test body-forms))
(defmacro do*-build-list (bindings test &rest body-forms)
  (loop-build-list 'do* bindings test body-forms))
(defmacro do1-build-list (bindings test &rest body-forms)
  (loop-build-list 'do1 bindings test body-forms))
(defmacro do1*-build-list (bindings test &rest body-forms)
  (loop-build-list 'do1* bindings test body-forms))

;; add to .emacs
;; (put 'do1 'lisp-indent-function 2)
;; (put 'do1* 'lisp-indent-function 2)
;; (put 'do-build-list 'lisp-indent-function 2)
;; (put 'do*-build-list 'lisp-indent-function 2)
;; (put 'do1-build-list 'lisp-indent-function 2)
;; (put 'do1*-build-list 'lisp-indent-function 2)
;; (defun my-elisp-mode-keywords()
;;   (font-lock-add-keywords 'lisp-mode
;;     '(
;;       ("(\\(\\<do1\\>\\)" . 1)
;;       ("(\\(\\<do1\\*\\>\\)" . 1)

;;       ("(\\(\\<do-build-list\\>\\)" . 1)
;;       ("(\\(\\<do1-build-list\\>\\)" . 1)
;;       ("(\\(\\<do\\*-build-list\\>\\)" . 1)
;;       ("(\\(\\<do1\\*-build-list\\>\\)" . 1)
;;     )
;;   )
;; )
;; (add-hook 'emacs-lisp-mode-hook 'my-elisp-mode-keywords)

(defun geometric-mean (x y) (sqrt (* x y)))

(defun cosine-period (srate freq)
  (snd-osc (first *sine-table*) (second *sine-table*)
	   srate freq 0.0 (/ 1.0 freq) 90.0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; find clicks at one frequency

;;; constructor and accessors
(defmacro click-detail (frequency steps ratio1 ratio2)
  `(vector ,frequency ,steps (geometric-mean ,ratio1 ,ratio2)))
(defmacro click-detail-frequency (detail) ; Hz
  `(aref ,detail 0))
(defmacro click-detail-steps (detail) ; integer to mulitply by step duration
  `(aref ,detail 1))
(defmacro click-detail-ratio (detail) ; float (dimensionless)
  `(aref ,detail 2))
(defmacro click-detail-difference (detail) ; float (dB)
  `(linear-to-db (aref ,detail 2)))

(labels
    ;; scan a sound until the end or the next sample meeting the criterion
    ;; (at or above the relative threshold, or at or below its reciprocal)
    ;; return the sample, or nil if end of sound found, or t if the limit
    ;; number of samples is scanned without finding it, paired with the number of
    ;; samples examined
    ((scan-sound
      (snd threshold limit)		; limit is inclusive
      (let* ((inv-threshold (/ threshold))
	     (nn 0)
	     (result (dotimes (ii limit t)
		       (let ((sample (snd-fetch snd)))
			 (cond (sample (inc nn)
				       (if (or (>= sample threshold)
					       (<= sample inv-threshold))
					   (return sample)))
			       (t (return nil)))))))
	(cons result nn)))

     (make-single-frequency-click-iterator
      (orig-srate amp-snd quotient-snd max-steps step-dur
		  relative-threshold absolute-threshold frequency)
      (let* ((t0 (snd-t0 amp-snd))
	     (my-quotient-snd (snd-copy quotient-snd))
	     ;; Recall that the first sample of quotient sound is later
	     ;; than that of the original by one block. Thus:
	     (nn (1- max-steps))
	     (sample-time (/ orig-srate))
	     start-val		   ; ratio at start of possible click,
					;or nil if not found yet
	     start)
	#'(lambda (limit-time)
	    ;; function may return t if the next click is at or after limit-time.
	    ;; limit-samples is inclusive limit for scanning new samples
	    (let ((limit-samples
		   (+ max-steps
		      (- nn)
		      (truncate (/ (- limit-time t0 sample-time) step-dur)))))
	      (loop
	       (let* ((result
		       (scan-sound
			my-quotient-snd relative-threshold limit-samples))
		      (new-val (car result)))
		 (add-to nn (cdr result))
		 (add-to limit-samples (- (cdr result)))
		 (cond
		  ((not (numberp new-val)) (return new-val)) ; nil or t
		  ((>= new-val relative-threshold)
		   ;; new possible start of click
		   (setf start-val new-val
			 start nn))
		  ((not start-val)
		   ;; step down found before step up
		   )
		  ((or
		    (> (- nn start) max-steps)
		    (< (snd-sref amp-snd (+ t0 (* start step-dur)))
		       absolute-threshold))
		   ;; step down found, but
		   ;; region is too long or too quiet
		   (setf start-val nil))
		  (t
		   ;; we have a winner
		   (return
		    (prog1 (cons (+ t0 (* start step-dur))
				 (click-detail frequency (- nn start)
					       start-val (/ new-val)))
		      (setf start-val nil))))))))))))

;;; returns iterator that returns (time . click-detail) or nil or t
;;; The iterator takes a limiting time argument.  If no click is before that
;;; time but input remains, the result is t.
;;; (This is done solely to improve progress indicator behavior.)
  (defun click-detail-finder (snd frequency steps-per-block step-size
				  relative-threshold absolute-threshold)
    (let* ((srate (snd-srate snd))
	   (step-dur (/ step-size srate))
	   (block-size (* steps-per-block step-size))
	   ;; the coefficient is such that convolving the window with
	   ;; a sinusoid of the same frequency leaves amplitude unchanged.
	   (window (prod (* -2.0 frequency (/ srate))
			 (cosine-period srate frequency)))
	   ;; convolve snd with window;
	   ;; discard some initial samples, so that the first remaining sample
	   ;; centers the window at the first sample of snd; that should
	   ;; leave the phase of a sinusoid of frequency Hz also unchanged.
	   (convolution (convolve snd window))
	   (shifted-convolution
	    (snd-xform convolution
		       (snd-srate convolution) (snd-t0 convolution)
		       (+ (snd-t0 convolution) (* 0.5 (/ frequency)))
		       MAX-STOP-TIME 1.0))
	   ;; find peak amplitude per block
	   (sig
	    (sum (db-to-linear -100)  ;guard against divisions by zero
		 (snd-avg shifted-convolution block-size step-size OP-PEAK)))
	   ;; now find the ratio of each peak to the peak one block before
	   ;; there will be no data for the first block of snd
	   (denominator (shift-time sig (/ block-size srate)))
	   (quotient-snd (prod sig (recip denominator))))
      (make-single-frequency-click-iterator
       srate sig quotient-snd steps-per-block step-dur relative-threshold
       absolute-threshold frequency))))

(flet
    ((consume-dummies
      (iter-arr lookaheads limit)
      (dotimes (ii (length lookaheads))
	(let ((val (aref lookaheads ii)))
	  (if (and val (not (consp val)))
	      (setf (aref lookaheads ii)
		    (funcall (aref iter-arr ii) limit)))))))

;;; First argument is array of iterators, each for a different frequency.
;;; The iterators return increasing start times.
;;; The output returns nondecrseasing start times, with a minor sort according
;;; to the input array.  Destroys iter-arr.
  (defun interleave-click-detail-finders (iter-arr t0 step-dur limit-time-step)
    (let* ((nn (length iter-arr))
	   (lookaheads (make-array nn))
	   (limit-time (+ t0 limit-time-step))
	   least
	   (position nn))
      (dotimes (ii nn)
	(setf (aref lookaheads ii)
	      (funcall (aref iter-arr ii) limit-time)))
      #'(lambda ()
	  (loop
	   (let ((answer (do ()
			     ((= position nn) nil)
			   (let ((val (aref lookaheads position)))
			     (cond ((and val (consp val)
					 (= (car val) least))
				    (setf (aref lookaheads position)
					  (funcall (aref iter-arr position)
						   limit-time))
				    (inc position)
				    (return val))
				   (t (inc position)))))))
	     (if answer (return answer))
	     ;; find the least time, but don't advance the iterators too
	     ;; unequally in time, so that the progress indicator behaves
	     ;; better.
	     (setf least nil)
	     (let (more)
	       (dotimes (ii nn)
		 (let ((val (aref lookaheads ii)))
		   (cond ((not val))
			 ((not (consp val)) (setf more t))
			 ((or (not least) (< (car val) least))
			  (setf least (car val))))))
	       (cond
		((and least (< least limit-time)) (setf position 0))
		(more (consume-dummies iter-arr lookaheads
				       (add-to limit-time limit-time-step)))
		(least (setf limit-time (+ least (/ step-dur 2))  position 0))
		(t (return nil))))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; find clicks in ranges of frequencies

;;; constructor and accessors
(defmacro click-info (start steps low high details &optional extra)
  `(vector ,start ,steps ,low ,high ,details ,extra))
(defmacro click-info-start (info) ; global time
  `(aref ,info 0))
(defmacro click-info-steps (info) ; integer to multiply by step duration
  `(aref ,info 1))
(defmacro click-info-low (info) ; Hz
  `(aref ,info 2))
(defmacro click-info-high (info) ; Hz
  `(aref ,info 3))
(defmacro click-info-details (info) ; list of click-detail
  `(aref ,info 4))
(defmacro click-info-extra (info) ; anything
  `(aref ,info 5))

(defun click-details-max-gain (details) ; returns dB
  (let (result)
    (dolist (detail details (linear-to-db result))
      (update-max result (click-detail-ratio detail)))))

(defun click-info-max-gain (info) ; returns dB
  (click-details-max-gain (click-info-details info)))

;;; Given an array of iterators returning (time . click-detail)
;;; at nondecreasing start times.
;;; Return an iterator returning combined click information for
;;; nondecreasing start times (increasing if combine-unequal-lengths is true).
;;; Only combines details for the same starting time.
;;; Does not combine all overlapping cases.
(defun combine-click-details (iter tag combine-unequal-lengths)
  (let ((lookahead (funcall iter)))
    #'(lambda ()
	(and lookahead
	     (let* ((time (car lookahead))
		    (detail (cdr lookahead))
		    (steps (click-detail-steps detail))
		    (low (click-detail-frequency detail))
		    (high low)
		    (details
		     (cons
		      (cdr lookahead)
		      (do1-build-list
			  ((_ (setf lookahead (funcall iter))))
			  (not (and lookahead
				    (= time (car lookahead))
				    (or combine-unequal-lengths
					(= steps (click-detail-steps
						  (cdr lookahead))))))
			(let ((detail (cdr lookahead)))
			  (update-min low (click-detail-frequency detail))
			  (update-max high (click-detail-frequency detail))
			  (update-max steps (click-detail-steps detail))
			  detail)))))
	       (click-info time steps low high details tag))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Takes an iterator returning click-info with nondecreasing start times,
;;; returns an iterator returning lists of click-info,
;;; grouping clicks that overlap in time, when their intervals are extended
;;; at each end by the amount extend-interval.
;;; (i.e. each group is a class of the equivalence closure of the overlapping
;;; relation.)

;; This function is not used in this file but may be useful if you want
;; to adapt the file for effects.
(defun group-overlapping-clicks (click-iterator step-dur extend-interval)
  (let ((lookahead (funcall click-iterator)))
    #'(lambda ()
	(and lookahead
	     (let* ((start (- (click-info-start lookahead) extend-interval))
		    (end (+ start (* step-dur (click-info-steps lookahead))
			    extend-interval)))
	       (cons lookahead
		     (do1-build-list
			 ((_ (setf lookahead (funcall click-iterator)))
			  (next-start (and lookahead
					   (click-info-start lookahead))))
			 (or (not next-start)
			     (<= end (- next-start extend-interval)))
		       (update-max end
				   (+ next-start
				      (* step-dur (click-info-steps lookahead))
				      extend-interval))
		       lookahead)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; returns iterator returning click-info for nondecreasing start times
;;; (increasing if combine-unequal-lengths is true)
;;; limit-time-step argument affects progress indicator smoothness only
(defun make-click-iterator (snd frequency-list
				steps-per-block step-size
				relative-threshold absolute-threshold tag
				&optional (limit-time-step MAX-STOP-TIME)
				(combine-unequal-lengths t))
  (let* ((click-detail-finders
	  (apply #'vector
		 (do-build-list
		     ((lis frequency-list (cdr lis)))
		     (not lis)
		   (click-detail-finder
		    snd (car lis) steps-per-block step-size
		    relative-threshold absolute-threshold))))
	 (interleaved-finder
	  (interleave-click-detail-finders
	   click-detail-finders (snd-t0 snd) (/ step-size (snd-srate snd))
	   limit-time-step)))
    (combine-click-details interleaved-finder tag combine-unequal-lengths)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; convert click-info to labels
;;; If the extra slot of a click is not an atom, assume it is another
;;; list of details for the right channel, with the same times and frequencies
(defun make-labels (click-iterator step-dur)
  (do1-build-list
      ((click (funcall click-iterator)))
      (not click)
    (let* ((time (click-info-start click))
	   (steps (click-info-steps click))
	   (low-freq (click-info-low click))
	   (high-freq (click-info-high click))
	   (gain (click-info-max-gain click))
	   (tag (click-info-extra click))
	   (gain2 (and (not (atom tag))
		       (click-details-max-gain tag)))
	   (prefix-string
	    (cond
	     ((not (atom tag)) "L/R:")
	     ((eq tag 'L) "L:")
	     ((eq tag 'R) "R:")
	     (t "")))
	   (freq-string
	    (if (= low-freq high-freq)
		(format nil "~A" low-freq)
		(format nil "~A-~A" low-freq high-freq)))
	   (db-string
	    (if (and gain2 (not (= gain gain2)))
		(format nil "~A/~A" gain gain2)
		(format nil "~A" gain)))
	   (result-string
	    (format nil "~A~A Hz (~A dB)" prefix-string freq-string db-string)))
      (list time (+ time (* steps step-dur)) result-string))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Main routines use the global variables

(defun make-channel-results (snd step-size tag)
  (let* ((frequency-list (do-build-list
			     ((frequency control-frequency-floor
				    (+ frequency control-frequency-step)))
			     (> frequency control-frequency-ceiling)
			   frequency)))
    (make-click-iterator
     snd frequency-list control-steps-per-block step-size
     (db-to-linear control-relative-threshold)
     (db-to-linear control-absolute-threshold) tag
     (/ (get-duration 1) 100.0))))

(defun make-mono-results (snd step-size)
  (let ((channel-results (make-channel-results snd step-size nil)))
    (make-labels channel-results (/ step-size (snd-srate snd)))))

(labels 
    ((compare (num1 num2)
	      (cond
	       ((< num1 num2) -1)
	       ((< num2 num1) 1)
	       (t nil)))

     ;; compare start time and low frequency before the others, to
     ;; agree with the sort order of the two streams.  Then compare
     ;; the rest, when deciding whether to merge.
     (compare-clicks
      (click1 click2)
      (or
       (compare (click-info-start click1) (click-info-start click2))
       (compare (click-info-low click1) (click-info-low click2))
       (compare (click-info-steps click1) (click-info-steps click2))
       (compare (click-info-high click1) (click-info-high click2))
       0))

;;; Take clicks from two channels, and wherever two have the same times
;;; and frequencies (but maybe not the same in other info), merge them.
     (merge-stereo-results
      (iter1 iter2)
      (let ((click1 (funcall iter1))
	    (click2 (funcall iter2)))
	#'(lambda ()
	    (and (or click1 click2)
		 (let ((choice (cond ((not click2) -1)
				     ((not click1) 1)
				     (t (compare-clicks click1 click2)))))
		   (case choice
			 (-1 (prog1 click1 (setf click1 (funcall iter1))))
			 (1 (prog1 click2 (setf click2 (funcall iter2))))
			 (0 (prog1
				;; construct a new click with right channel
				;; details in the extra slot
				(click-info
				 (click-info-start click1)
				 (click-info-steps click1)
				 (click-info-low click1)
				 (click-info-high click1)
				 (click-info-details click1)
				 (click-info-details click2))
			      (setf click1 (funcall iter1)
				    click2 (funcall iter2)))))))))))

  (defun make-stereo-results (snds step-size)
    (let ((channel-results-L (make-channel-results (aref snds 0) step-size 'L))
	  (channel-results-R (make-channel-results (aref snds 1) step-size 'R)))
      (if (soundp channel-results-L)
	  (vector channel-results-L channel-results-R)
	  (make-labels
	   (merge-stereo-results channel-results-L channel-results-R)
	   (/ step-size (snd-srate (aref snds 0))))))))

;;; deal with mono or stereo inputs
(let* ( ;; throughout the rest, the actual minimum separation will be
       ;; step-size times control-steps-per-block in samples,
       ;; which might be a few samples different from control-separation.
       ;; I don't care.
       (srate (snd-srate (if (arrayp s) (aref s 0) s)))
       (step-size
	(truncate (* srate (/ control-separation
			      (* 1000.0 control-steps-per-block))))))
  (or
   (if (arrayp s)
       (make-stereo-results s step-size)
       (make-mono-results s step-size))
   "Found no labels"))
